from spikeextractors import RecordingExtractor, SortingExtractor
import numpy as np


class OpenEphysRecordingExtractor(RecordingExtractor):
    def __init__(self, recording_file, *, probefile=None, experiment_id=0, recording_id=0):
        try:
            import pyopenephys
        except ModuleNotFoundError:
            raise ModuleNotFoundError("To use the OpenEphys extractor, install pyopenephys: \n\n"
                                      "pip install pyopenephys\n\n")
        RecordingExtractor.__init__(self)
        self._recording_file = recording_file
        self._recording = pyopenephys.File(recording_file,
                                           probefile).experiments[experiment_id].recordings[recording_id]

    def getChannelIds(self):
        return list(range(self._recording.analog_signals[0].signal.shape[0]))

    def getNumFrames(self):
        return self._recording.analog_signals[0].signal.shape[1]

    def getSamplingFrequency(self):
        return float(self._recording.sample_rate.rescale('Hz').magnitude)

    def getTraces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.getNumFrames()
        if channel_ids is None:
            channel_ids = range(self.getNumChannels())

        return self._recording.analog_signals[0].signal[channel_ids, start_frame:end_frame]


class OpenEphysSortingExtractor(SortingExtractor):
    def __init__(self, recording_file, *, probefile=None, experiment_id=0, recording_id=0):
        try:
            import pyopenephys
        except ModuleNotFoundError:
            raise ModuleNotFoundError("To use the OpenEphys extractor, install pyopenephys: \n\n"
                                      "pip install pyopenephys\n\n")
        SortingExtractor.__init__(self)
        self._recording_file = recording_file
        self._recording = pyopenephys.File(recording_file,
                                           probefile).experiments[experiment_id].recordings[recording_id]
        self._spiketrains = self._recording.spiketrains
        self._unit_ids = list([np.unique(st.clusters)[0] for st in self._spiketrains])

    def getUnitIds(self):
        return self._unit_ids

    def getUnitSpikeTrain(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        st = self._spiketrains[unit_id]
        inds = np.where((start_frame <= (st.times * self._recording.sample_rate)) &
                        ((st.times * self._recording.sample_rate) < end_frame))
        return (st.times[inds] * self._recording.sample_rate).magnitude



from .SortingExtractor import SortingExtractor
import numpy as np


# Encapsulates a subset of a spike sorted data file

class SubSortingExtractor(SortingExtractor):

    def __init__(self, parent_sorting, *, unit_ids=None, renamed_unit_ids=None, start_frame=None, end_frame=None):
        SortingExtractor.__init__(self)
        self._parent_sorting = parent_sorting
        self._unit_ids = unit_ids
        self._renamed_unit_ids = renamed_unit_ids
        self._start_frame = start_frame
        self._end_frame = end_frame
        if self._unit_ids is None:
            self._unit_ids = self._parent_sorting.getUnitIds()
        if self._renamed_unit_ids is None:
            self._renamed_unit_ids = self._unit_ids
        if self._start_frame is None:
            self._start_frame = 0
        if self._end_frame is None:
            self._end_frame = float("inf")
        self._original_unit_id_lookup = {}
        for i in range(len(self._unit_ids)):
            self._original_unit_id_lookup[self._renamed_unit_ids[i]] = self._unit_ids[i]

    def getUnitIds(self):
        return self._renamed_unit_ids

    def getUnitSpikeTrain(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        if (isinstance(unit_id, (int, np.integer))):
            if (unit_id in self.getUnitIds()):
                original_unit_id = self._original_unit_id_lookup[unit_id]
            else:
                raise ValueError("Non-valid channel_id")
        else:
            raise ValueError("channel_id must be an int")
        sf = self._start_frame + start_frame
        ef = self._start_frame + end_frame
        if sf < self._start_frame:
            sf = self._start_frame
        if ef > self._end_frame:
            ef = self._end_frame
        return self._parent_sorting.getUnitSpikeTrain(unit_id=original_unit_id, start_frame=sf,
                                                      end_frame=ef) - self._start_frame



from spikeextractors import SortingExtractor
import numpy as np
from pathlib import Path


class PhySortingExtractor(SortingExtractor):
    def __init__(self, phy_folder):
        SortingExtractor.__init__(self)
        phy_folder = Path(phy_folder)

        spike_times = np.load(phy_folder / 'spike_times.npy')
        spike_templates = np.load(phy_folder / 'spike_templates.npy')
        amplitudes = np.load(phy_folder / 'amplitudes.npy')
        pc_features = np.load(phy_folder / 'pc_features.npy')

        if (phy_folder /'spike_clusters.npy').is_file():
            spike_clusters = np.load(phy_folder / 'spike_clusters.npy')
        else:
            spike_clusters = spike_templates

        self._spiketrains = []
        self._amps = []
        self._pc_features = []
        clust_id = np.unique(spike_clusters)
        self._unit_ids = list(clust_id)
        spike_times.astype(int)

        self._spiketrais = []
        for clust in self._unit_ids:
            idx = np.where(spike_clusters == clust)[0]
            self._spiketrains.append(spike_times[idx])
            self._amps.append(amplitudes[idx])
            self._pc_features.append(pc_features[idx])

        # set features
        for u_i, unit in enumerate(self.getUnitIds()):
            self.setUnitSpikeFeatures(unit, 'amplitudes', self._amps[u_i])
            self.setUnitSpikeFeatures(unit, 'pc_features', self._pc_features[u_i])

    def getUnitIds(self):
        return list(self._unit_ids)

    def getUnitSpikeTrain(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        times = self._spiketrains[self.getUnitIds().index(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]

    @staticmethod
    def writeSorting(sorting, save_path):
        save_path = Path(save_path)
        spike_times = np.array([])
        spike_clusters = np.array([])
        amplitudes = np.array([])
        pc_features = np.array([])

        for id in sorting.getUnitIds():
            st = sorting.getUnitSpikeTrain(id)
            cl = [id] * len(sorting.getUnitSpikeTrain(id))
            spike_times = np.concatenate((spike_times, np.array(st)))
            spike_clusters = np.concatenate((spike_clusters, np.array(cl)))
            if 'amplitudes' in sorting.getUnitSpikeFeatureNames():
                amp = sorting.getUnitSpikeFeatures(id, 'amplitudes')
                amplitudes = np.concatenate((amplitudes, np.array(amp)))
            if 'pc_features' in sorting.getUnitSpikeFeatureNames():
                pc_feat = sorting.getUnitSpikeFeatures(id, 'pc_features')
                if len(pc_features) == 0:
                    pc_features = pc_feat
                else:
                    pc_features = np.vstack((pc_features, np.array(pc_feat)))

        sorting_idxs = np.argsort(spike_times)
        spike_times = spike_times[sorting_idxs]
        spike_clusters = spike_clusters[sorting_idxs]

        if not save_path.is_dir():
            save_path.mkdirs()
        np.save(save_path /'spike_times.npy', spike_times[:, np.newaxis].astype(int))
        np.save(save_path / 'spike_clusters.npy', spike_clusters[:, np.newaxis].astype(int))
        if len(amplitudes) > 0:
            amplitudes = amplitudes[sorting_idxs]
            np.save(save_path / 'amplitudes.npy', amplitudes[:, np.newaxis].astype(int))
        if len(pc_features) > 0:
            pc_features = pc_features[sorting_idxs]
            np.save(save_path / 'pc_features.npy', pc_features)
            pc_feature_ind = np.tile(np.arange(pc_features.shape[-1]), (len(sorting.getUnitIds()), 1))
            np.save(save_path / 'pc_feature_ind.npy', pc_feature_ind)



from .RecordingExtractor import RecordingExtractor
from .SortingExtractor import SortingExtractor
from .SubSortingExtractor import SubSortingExtractor
from .SubRecordingExtractor import SubRecordingExtractor
from .MultiRecordingExtractor import MultiRecordingExtractor
from .MultiSortingExtractor import MultiSortingExtractor
from .CuratedSortingExtractor import CuratedSortingExtractor

from .extractors.mdaextractors.mdaextractors import MdaRecordingExtractor, MdaSortingExtractor
from .extractors.mearecextractors.mearecextractors import MEArecRecordingExtractor, MEArecSortingExtractor
from .extractors.biocamrecordingextractor import BiocamRecordingExtractor
from .extractors.exdirextractors import ExdirRecordingExtractor, ExdirSortingExtractor
from .extractors.hs2sortingextractor import HS2SortingExtractor
from .extractors.klustasortingextractor import KlustaSortingExtractor
from .extractors.kilosortsortingextractor import KiloSortSortingExtractor
from .extractors.numpyextractors.numpyextractors import NumpyRecordingExtractor, NumpySortingExtractor
from .extractors.nwbextractors.nwbextractors import NwbRecordingExtractor
from .extractors.openephysextractors.openephysextractors import OpenEphysRecordingExtractor, OpenEphysSortingExtractor
from .extractors.physortingextractor.physortingextractor import PhySortingExtractor
from .extractors.rawrecordingextractor import RawRecordingExtractor
from .extractors.spykingcircussortingextractor.spykingcircussortingextractor import SpykingCircusSortingExtractor

from . import example_datasets
from .tools import loadProbeFile, saveProbeFile, writeBinaryDatFormat, getSubExtractorsByProperty
